/*******************************************************************************
 *
 * blueMarine - open source photo workflow
 * =======================================
 *
 * Copyright (C) 2003-2010 by Fabrizio Giudici
 * Project home page: http://bluemarine.tidalwave.it
 *
 *******************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *******************************************************************************
 *
 * $Id: ElmoFinder.java,v 1c263a6d8ac9 2010/02/14 00:23:33 fabrizio $
 *
 ******************************************************************************/
package it.tidalwave.openrdf.elmo;

import javax.annotation.Nonnegative;
import javax.annotation.Nonnull;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeSet;
import javax.xml.namespace.QName;
import org.openrdf.elmo.ElmoManager;
import org.openrdf.elmo.ElmoQuery;
import it.tidalwave.util.logging.Logger;
import it.tidalwave.util.Finder;
import it.tidalwave.util.NotFoundException;
import it.tidalwave.semantic.Wrapper;

/*******************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id: ElmoFinder.java,v 1c263a6d8ac9 2010/02/14 00:23:33 fabrizio $
 *
 ******************************************************************************/
public abstract class ElmoFinder<T> implements Finder<T>
  {
    public static interface QueryFactory
      {
        @Nonnull
        public ElmoQuery createQuery (@Nonnull String queryString);
      }

    private static final String CLASS = ElmoFinder.class.getName();
    private static final Logger logger = Logger.getLogger(CLASS);

    private static final String QUERY ="PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>\n" +
                                       "%s" +
                                       "SELECT ?%s\n" +
                                       "WHERE\n" +
                                       "  {\n" +
                                       "    ?%s a ?type.\n" +
                                       "%s" +
                                       "  }\n" +
                                       "%s";

    // Can be replaced for testing purposes only
    public static QueryFactory queryFactory = new QueryFactory()
      {
        private final ElmoManager em = new ElmoManagerProxy();

        @Override
        @Nonnull
        public ElmoQuery createQuery (final @Nonnull String queryString)
          {
            return em.createQuery(queryString);
          }
      };

    protected int firstResult = 0;

    protected int maxResults = Integer.MAX_VALUE - 1;

    protected final List<SortCriterion> sortCriteria = new ArrayList<SortCriterion>();

    protected final List<SortDirection> sortDirections = new ArrayList<SortDirection>();

    @Nonnull
    protected String queryString = "";
    
    @Nonnull
    protected Class<T> entityClass;

    @Nonnull
    protected Class<? extends T> entityImplClass;

    @Nonnull
    protected final String nameSpaceHeader;

    protected String entityName;

    protected final Map<String, QName> idMapByPlaceHolder = new HashMap<String, QName>();

    protected final Set<String> auxStatements = new TreeSet<String>();

    /***************************************************************************
     *
     *
     **************************************************************************/
    public ElmoFinder()
      {
        entityClass = null;
        entityImplClass = null;
        nameSpaceHeader = null;
        entityName = null;
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    public ElmoFinder (final @Nonnull String nameSpaceHeader)
      {
        this.nameSpaceHeader = nameSpaceHeader;
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    public ElmoFinder (final @Nonnull Class<T> entityClass,
                       final @Nonnull Class<? extends T> entityImplClass,
                       final @Nonnull String nameSpaceHeader)
      {
        this.entityClass = entityClass;
        this.entityImplClass = entityImplClass;
        this.nameSpaceHeader = nameSpaceHeader;
        setEntityName();
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    @Nonnull
    @Override
    public Finder<T> from (final @Nonnegative int firstResult)
      {
        this.firstResult = firstResult;
        return this;
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    @Nonnull
    @Override
    public Finder<T> max (final @Nonnegative int maxResults)
      {
        this.maxResults = maxResults;
        return this;
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    @Nonnull
    @Override
    public <X> Finder<X> ofType (final @Nonnull Class<X> type)
      {
        entityClass = (Class<T>)type;
        entityImplClass = (Class<? extends T>)Wrapper.findImplementation(type);

        if (entityImplClass == null)
          {
            throw new IllegalArgumentException(String.format("No implementation for class %s", type.getName()));
          }

        setEntityName();
        return (Finder<X>)this;
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    @Nonnull
    @Override
    public Finder<T> sort (final @Nonnull SortCriterion criterion, final @Nonnull SortDirection direction)
      {
        sortCriteria.add(criterion);
        sortDirections.add(direction);
        return this;
      }

    
    /***************************************************************************
     *
     *
     **************************************************************************/
    @Nonnull
    @Override
    public Finder<T> sort (final @Nonnull SortCriterion criterion)
      {
        return sort(criterion, SortDirection.ASCENDING);
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    @Override
    @Nonnegative
    public int count()
      {
        logger.fine("count()");
        final ElmoQuery query = createQuery();
        query.setFirstResult(firstResult).setMaxResults(maxResults);
        final Integer result = query.getResultList().size();
        // FIXME: no count(*) in SPARQL :-((
        // See http://www.openrdf.org/issues/browse/SES-411
//        final ElmoQuery query = createQuery("count (?object)", false);
//        final Integer result = (Integer)query.getSingleResult();
        logger.finer(">>>> returning %d", result);

        return result;
      }


    /***************************************************************************
     *
     * {@inheritDoc}
     *
     **************************************************************************/
    @Nonnull
    @Override
    public List<? extends T> results()
      {
        final ElmoQuery query = createQuery();
        logger.fine(">>>> limit: first: %d, max: %d", firstResult, maxResults);
        query.setFirstResult(firstResult).setMaxResults(maxResults);
        final List results = Wrapper.wrap(query.getResultList());
        logger.finer(">>>>results: %s", results);
        return results;
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    @Nonnull
    @Override
    public final T result()
      throws NotFoundException
      {
        final List<? extends T> results = results();

        switch (results.size())
          {
            case 0: throw new NotFoundException(getNameForException());
            case 1: return results.get(0);
            default: throw new RuntimeException(getNameForException() + "Unexpected multiple results. " + results);
          }
      }

    @Nonnull
    private String getNameForException()
      {
        return String.format("%s\ntype: %s\nids: %s\nfirst: %d, max: %d\n",
                             queryString, entityImplClass.getName(), idMapByPlaceHolder, firstResult, maxResults);
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    @Nonnull
    @Override
    public final T firstResult()
      throws NotFoundException
      {
        final List<? extends T> results = results();

        if (results.isEmpty())
          {
            throw new NotFoundException(getNameForException());
          }

        return results.get(0);
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    @Nonnull
    @Override
    public Finder<T> clone()
      {
        final ElmoFinder<T> clone = (ElmoFinder<T>)createClone();
        clone.firstResult = this.firstResult;
        clone.maxResults = this.maxResults;
        clone.sortCriteria.addAll(sortCriteria);
        clone.sortDirections.addAll(sortDirections);
        return clone;
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    @Nonnull
    protected ElmoQuery createQuery()
      {
        return createQuery(entityName, true);
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    @Nonnull
    protected ElmoQuery createQuery (final @Nonnull String entityName, final boolean sort)
      {
        if (entityImplClass == null)
          {
            throw new IllegalArgumentException("Illegal class " + entityClass);
          }

        final Set<String> auxStatements = new TreeSet<String>(this.auxStatements);
        final StringBuilder orderBy = new StringBuilder();

        process(auxStatements);

        if (sort)
          {
            orderBy.append(sortCriteria.isEmpty() ? "" : "ORDER BY ");
            String separator = "";

            for (int i = 0; i < sortCriteria.size(); i++)
              {
                final SortCriterion criterion = (SortCriterion)sortCriteria.get(i);
                final SortDirection direction = sortDirections.get(i);
                orderBy.append(separator);
                orderBy.append((direction == SortDirection.DESCENDING) ? "DESC(" : "ASC(");
                processSortCriterion(criterion, orderBy, auxStatements);
                orderBy.append(")");
                separator = ", ";
              }
          }

        final StringBuilder auxStatementsToString = new StringBuilder();

        for (final String string : auxStatements)
          {
            auxStatementsToString.append("    ").append(string).append("\n");
          }

        queryString = String.format(QUERY, nameSpaceHeader, entityName, entityName, auxStatementsToString.toString(), orderBy.toString());
        logger.fine(">>>> query: %s", queryString);
        logger.fine(">>>> type:  %s", entityImplClass.getName());
        logger.fine(">>>> ids:   %s", idMapByPlaceHolder);
        final ElmoQuery query = queryFactory.createQuery(queryString).setType("type", entityImplClass);

        for (final Entry<String, QName> entry : idMapByPlaceHolder.entrySet())
          {
            query.setQName(entry.getKey(), entry.getValue());
          }

        return query;
      }
    
    /***************************************************************************
     *
     *
     **************************************************************************/
    @Nonnull
    protected abstract Finder<T> createClone();

    /***************************************************************************
     *
     *
     **************************************************************************/
    protected void process (final @Nonnull Set<String> auxStatements)
      {
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    protected void processSortCriterion (final @Nonnull ElmoFinder.SortCriterion criterion,
                                         final @Nonnull StringBuilder orderBy,
                                         final @Nonnull Set<String> auxStatements)
      {
      }

    private void setEntityName()
      {
        final String s = entityClass.getName().replaceAll(".*\\.", "");
        entityName = s.substring(0, 1).toLowerCase() + s.substring(1);
      }
  }
